import Foundation
import AVFoundation
import SwiftUI
import React

extension AVAudioTime{
    var toSeconds:Double{
        Double(self.sampleTime)/self.sampleRate
    }
}
enum TremoloControlType {
    case WetDryMix
    case lowPassCutoff
    case feedback
}

@objc(NativeAudioEngine)
class NativeAudioEngine: RCTEventEmitter {
    
    @objc(multiply:withB:withResolver:withRejecter:)
    func multiply(a: Float, b: Float, resolve:RCTPromiseResolveBlock,reject:RCTPromiseRejectBlock) -> Void {
        resolve(a*b)
    }
    @objc override static func requiresMainQueueSetup()->Bool{ true }
    override func supportedEvents() -> [String]! {
        ["TimeChange","FrequencyEmitter","FinishPlaying"]
    }
    
    private var engine:AVAudioEngine!
//    private var mixer:AVAudioMixerNode!
    private var outputMixer:AVAudioMixerNode =  AVAudioMixerNode()
    private var inputMixer:AVAudioMixerNode =  AVAudioMixerNode()
    
    private var playerNode:AVAudioPlayerNode = AVAudioPlayerNode()
    private var tracks:[Track] = []
    private var fileExternal :AVAudioFile!
//    private var bufferPlayer :AVAudioPCMBuffer!
//    private var bufferRecorder :AVAudioPCMBuffer!
//    private var nodes:[AVAudioUnit] = []

    private var fileCurrentRecording : AVAudioFile? = nil
    
    private let filename = FileManager.default.urls(for: .cachesDirectory, in: .userDomainMask).first!.appendingPathComponent("file.m4a")
    private let urlCurrentRecordingFile = FileManager.default.urls(for: .cachesDirectory, in: .userDomainMask).first!.appendingPathComponent("temp.m4a")
    
    private var pausePosition:Double = 0
    private var currentPosition:Double = 0
//    private var currentRecordBuffer = [AVAudioPCMBuffer]()
    private let processQueue = DispatchQueue(label: "Audio.process",qos: .userInitiated,attributes: DispatchQueue.Attributes.concurrent)
    
    private var uiUpdater:Timer!
    private var liveOutputState = false
    
    var selectedTrack:Int = 0
    
    private let preverbPresets = [AVAudioUnitReverbPreset.smallRoom,AVAudioUnitReverbPreset.largeRoom,AVAudioUnitReverbPreset.largeHall]
    
    override init() {
        super.init()
        //setup session
        let session = AVAudioSession.sharedInstance()
        try! session.setCategory(.playAndRecord,mode: .default,options: [.allowBluetooth,.mixWithOthers,.defaultToSpeaker])

        try! session.setActive(true,options: .notifyOthersOnDeactivation)

        self.start()
     
    }
    private func start(){
        self.engine = AVAudioEngine()
        self.engine.attach(self.inputMixer)
        self.engine.attach(self.outputMixer)
//        self.engine.attach(self.mixer)
        self.engine.attach(self.playerNode)
        self.addTrack() // added an track on init process
        self.engine.attach(self.tracks[selectedTrack].playerNode)
        
        let inputFormmat = engine.inputNode.inputFormat(forBus: 0)
       
        let mixerNode = engine.mainMixerNode
        let mixerFormmat = AVAudioFormat(commonFormat: .pcmFormatFloat32, sampleRate: inputFormmat.sampleRate, channels: AVAudioChannelCount(1), interleaved: false)

        engine.connect(engine.inputNode, to: inputMixer, format:mixerFormmat)
        engine.connect(tracks[selectedTrack].playerNode, to: inputMixer, format: mixerFormmat)
        engine.connect(inputMixer, to: outputMixer, format: mixerFormmat)
        engine.connect(outputMixer, to: mixerNode, format: mixerFormmat)
        engine.connect(playerNode, to: mixerNode, format: mixerFormmat)
//        engine.connect(mixer, to: mixerNode, format: mixerFormmat)
        outputMixer.volume = 0
        try? FileManager.default.removeItem(at: urlCurrentRecordingFile)
        try? FileManager.default.removeItem(at: tracks[0].url)
        
        self.engine.prepare()
    }
    private func reset(){
        self.engine.stop()
        self.engine.reset()
        self.tracks.removeAll()
        self.currentPosition = 0
        self.pausePosition = 0
        self.sendEvent(withName: "TimeChange", body: 0)
    }
    private func startTimer(){
        DispatchQueue.main.async {
            self.uiUpdater = .scheduledTimer(timeInterval: 0.1, target: self, selector: #selector(self.uiUpdaterHandler), userInfo: nil, repeats: true)
        }
    }
    
    private func stopTimer(){
        uiUpdater?.invalidate()
        uiUpdater = nil
    }
    @objc func setExternalAudio(path:String){
        do{
            fileExternal = try AVAudioFile(forReading:.init(fileURLWithPath: path))
            playerNode.scheduleFile(fileExternal, at: nil)
        }catch{
            
        }
    }
    @objc func startRecording(){
        engine.inputNode.volume = 1
        outputMixer.volume = 0
       
        let tapNode:AVAudioNode = engine.inputNode
        let format = engine.inputNode.outputFormat(forBus: 0)
        do{
            let file = try AVAudioFile(forWriting:urlCurrentRecordingFile, settings: format.settings,commonFormat: .pcmFormatFloat32, interleaved: true)
            tapNode.installTap(onBus: 0, bufferSize: 1024, format: format) { buffer, time in
                guard let channelData = buffer.floatChannelData else {
                    return
                }
//                          let channelDataValue = channelData.pointee
//                          let channelDataValueArray = stride(from: 0,
//                              to: Int(buffer.frameLength),
//                              by: buffer.stride)
//                              .map { channelDataValue[$0] }
//
//                          // 5
//                          let rms = sqrt(channelDataValueArray.map {
//                              return $0 * $0
//                          }
//                          .reduce(0, +) / Float(buffer.frameLength))
//
//                          let avgPower = 20 * log10(rms)
//                          let meterLevel = self.scaledPower(power: avgPower)
                try! file.write(from: buffer)
//                self.currentRecordBuffer.append(buffer)
                self.sendEvent(withName: "FrequencyEmitter", body: channelData[0].pointee * 40)
            }
            try engine.start()
            playerNode.play()
            startTimer()
        }catch{
            print("Engine start erro : ",error.localizedDescription)
        }
    }
    
    @objc func finishRecording(){
        stopTimer()
        engine.inputNode.removeTap(onBus: 0)
        engine.inputNode.volume = 0
     
        engine.pause()
//        if let s = try? filename.resourceValues(forKeys: [.fileSizeKey]) ,let bytes = s.fileSize{
//            print(Float(bytes)/1024.0/1024.0,"MB")
//        }
        let cp = pausePosition
        processQueue.async { [self] in
            self.writeAt(at:cp, from: urlCurrentRecordingFile, to: tracks[selectedTrack].url)
        }
        self.pausePosition = currentPosition
        self.tracks[selectedTrack].endPosition = self.pausePosition
    }
    
    @objc func playAudio(){
//        liveOutputState = outputMixer.volume == 1
        outputMixer.volume = 1
        engine.inputNode.volume = 0
        seekPlayer(currentPosition)
        resume()
    }
    
    @objc func stopPlaying() {
//        outputMixer.volume = liveOutputState ? 1 : 0
        tracks.forEach {track in
            track.playerNode.pause()
            track.playerNode.stop()
        }
        engine.pause()
        self.stopTimer()
    }
    
    @objc func seekPlayer(_ s:Double){
      tracks.forEach { node in
          node.playerNode.stop()
          node.playerNode.reset()
          if s < node.endPosition, let file = try? AVAudioFile(forReading: node.url){
          let pos = s*file.fileFormat.sampleRate
          let scheduleAt:AVAudioTime? = node.startPosition < s ?  nil : AVAudioTime(hostTime: mach_absolute_time(), sampleTime: AVAudioFramePosition(s*file.fileFormat.sampleRate), atRate: file.fileFormat.sampleRate)
              if pos < Double(file.length){
                  node.playerNode.scheduleSegment(file, startingFrame: AVAudioFramePosition(pos), frameCount: AVAudioFrameCount(Double(file.length)-pos), at: scheduleAt, completionHandler: nil)
              }
        }
      }
        if let fileExternal{
            let pos = s*fileExternal.fileFormat.sampleRate
            playerNode.stop()
            playerNode.reset()
            if pos < Double(fileExternal.length){
                playerNode.scheduleSegment(fileExternal, startingFrame: AVAudioFramePosition(pos), frameCount: AVAudioFrameCount(UInt64(Double(fileExternal.length)-pos)), at: nil, completionHandler: nil)
            }
        }
        self.currentPosition = s
        self.pausePosition = s
    }
    
    @objc func addTrack() {
        let totalTracks = tracks.count
        selectedTrack = totalTracks
        let node = AVAudioPlayerNode()
        engine.attach(node)
        engine.connect(node, to: inputMixer, format: nil)
        tracks.append(Track(url: FileManager.default.urls(for: .cachesDirectory, in: .userDomainMask).first!.appendingPathComponent("track_\(totalTracks).m4a"), startPosition: pausePosition, endPosition: .init(), playerNode: node))
        
    }
    
    @objc func removeTrack(trackIndex at:Int) {
        if selectedTrack == at{
            selectedTrack -= 1
        }
        let node = tracks.remove(at: at)
        node.playerNode.reset()
        engine.detach(node.playerNode)
        try? FileManager.default.removeItem(at: node.url)
    }
    
    private func writeAt(at:Double,from:URL,to:URL){
        // merge old and recent audio
        let timeNow = Date().timeIntervalSinceReferenceDate
        
        guard let currentFile = try? AVAudioFile(forReading: from) else{
            #if DEBUG
            print("No recorded audio file exist!")
            #endif
            return
        }
        let pcmBufferCurrent = AVAudioPCMBuffer(pcmFormat: currentFile.processingFormat, frameCapacity: AVAudioFrameCount(currentFile.length))!
        try! currentFile.read(into: pcmBufferCurrent)
        
        var leftBuffer:AVAudioPCMBuffer? = nil
        var rightBuffer:AVAudioPCMBuffer? = nil
        var emptyBuffer:AVAudioPCMBuffer? = nil
        var isPositionSmaller = false
        if let destinationFile = try? AVAudioFile(forReading: to){
            
            let mergFileLength = destinationFile.length
            
            let leftEndPosition:Int64 = Int64(at*currentFile.fileFormat.sampleRate)
            let rightStartPosition = leftEndPosition + currentFile.length
            isPositionSmaller = leftEndPosition <= destinationFile.length
            leftBuffer = AVAudioPCMBuffer(pcmFormat: currentFile.processingFormat, frameCapacity: isPositionSmaller ? AVAudioFrameCount(leftEndPosition) : AVAudioFrameCount(destinationFile.length))!
            if leftEndPosition > 0{
                try! destinationFile.read(into: leftBuffer!, frameCount: leftBuffer!.frameCapacity)
            }
            
            if !isPositionSmaller,let _emptyBuffer = AVAudioPCMBuffer(pcmFormat: destinationFile.fileFormat, frameCapacity: AVAudioFrameCount(leftEndPosition - destinationFile.length)){
                _emptyBuffer.frameLength = _emptyBuffer.frameCapacity
                emptyBuffer = _emptyBuffer
            }
            let rigthBufferCount = AVAudioFrameCount(max(0, mergFileLength-rightStartPosition))
            rightBuffer = AVAudioPCMBuffer(pcmFormat: currentFile.processingFormat , frameCapacity: rigthBufferCount)
            
            if rightStartPosition < mergFileLength{
                destinationFile.framePosition = rightStartPosition
                try! destinationFile.read(into: rightBuffer!, frameCount: rightBuffer!.frameCapacity)
            }
        }
        let destinationFile = try! AVAudioFile(forWriting: to, settings: currentFile.processingFormat.settings)
        if let leftBuffer{
            try! destinationFile.write(from: leftBuffer)
        }
        if let emptyBuffer{
            try! destinationFile.write(from: emptyBuffer)
        }
        try! destinationFile.write(from: pcmBufferCurrent)
        
        if let rightBuffer{
            try! destinationFile.write(from: rightBuffer)
        }
    }
    
    @objc func exportFile(_ resolve:RCTPromiseResolveBlock,rejecter reject:RCTPromiseRejectBlock){
        playerNode.reset()
        tracks.forEach { t in
            t.playerNode.reset()
        }
        engine.stop()
        engine.reset()
        
        let outputPath = FileManager.default.urls(for: .libraryDirectory, in: .userDomainMask).first!.appendingPathComponent("outputAudio.aac")
        guard let format: AVAudioFormat = tracks.first?.playerNode.outputFormat(forBus: 0) else{
            reject("0","No track availble in audio session or currupted",nil)
            return
        }
        var maxFileLenght:Int64 = 0
        
        do {
            // The maximum number of frames the engine renders in any single render call.
            let maxFrames: AVAudioFrameCount = 4096
            try engine.enableManualRenderingMode(.offline, format: format,
                                                 maximumFrameCount: maxFrames)
        } catch {
            reject("0","Audio Engine failed to start export session!",nil)
        }

        engine.attach(playerNode)
        tracks.forEach { t in
            engine.attach(t.playerNode)
            if let effectNode = t.effectNode{
                engine.attach(effectNode)
                engine.connect(t.playerNode, to: effectNode, format: format)
                engine.connect(effectNode, to: inputMixer, format: format)
            }else{
                engine.connect(t.playerNode, to: inputMixer, format: format)
            }
        }
        engine.connect(inputMixer, to: engine.mainMixerNode, format: format)
        engine.connect(playerNode, to: inputMixer, format:format )
        tracks.forEach { track in
            if let afile = try? AVAudioFile(forReading:track.url){
                if afile.length > maxFileLenght{
                    maxFileLenght = afile.length
                }
                track.playerNode.scheduleFile(afile, at: nil)
            }
        }
        if let fileExternal{
            if fileExternal.length > maxFileLenght{
                maxFileLenght = fileExternal.length
            }
            playerNode.scheduleFile(fileExternal, at: nil)
        }
        
        self.engine.prepare()
        do {
            try engine.start()
            playerNode.play()
            tracks.forEach { track in
                track.playerNode.play()
            }
        } catch {
            reject("0","Unable to start audio engine: error",nil)
        }
        
        let buffer = AVAudioPCMBuffer(pcmFormat: engine.manualRenderingFormat,
                                      frameCapacity: engine.manualRenderingMaximumFrameCount)!

        let outputFile: AVAudioFile
        do {
            outputFile = try AVAudioFile(forWriting: outputPath, settings:  [AVFormatIDKey:kAudioFormatMPEG4AAC,AVLinearPCMIsFloatKey: true],commonFormat: .pcmFormatFloat32,interleaved: false)
        } catch {
            fatalError("Unable to open output audio file: \(error).")
        }
        
        while engine.manualRenderingSampleTime < maxFileLenght {
            do {
                let frameCount = maxFileLenght - engine.manualRenderingSampleTime
                let framesToRender = min(AVAudioFrameCount(frameCount), buffer.frameCapacity)
                
                let status = try engine.renderOffline(framesToRender, to: buffer)
                
                switch status {
                case .success:
                    // The data rendered successfully. Write it to the output file.
                    try outputFile.write(from: buffer)
                    
                case .insufficientDataFromInputNode:
                    // Applicable only when using the input node as one of the sources.
                    break
                    
                case .cannotDoInCurrentContext:
                    // The engine couldn't render in the current render call.
                    // Retry in the next iteration.
                    break
                    
                case .error:
                    // An error occurred while rendering the audio.
                    reject("0","exporting audio render process unkown error",nil)
                @unknown default:
                   break
                }
            } catch {
                reject("0","exporting audio process throw unkown error",nil)
            }
        }
        playerNode.stop()
        
//         print(Date().timeIntervalSinceReferenceDate-timeNow)
        self.reset()
        self.start()
        resolve(outputPath.path)
    }
    
    private func scaledPower(power: Float) -> Float {
         guard power.isFinite else {
             return 0.0
         }
         let minDb: Float = -80
         if power < minDb {
             return 0.0
         } else if power >= 1.0 {
             return 1.0
         } else {
             return (abs(minDb) - abs(power)) / abs(minDb)
         }
     }
    @objc func pause(){
        self.stopTimer()
        self.engine.pause()
        self.playerNode.pause()
        self.tracks.forEach { t in
            t.playerNode.pause()
        }
    }
    
    @objc func resume(){
        do{
            try self.engine.start()
            self.startTimer()
            self.playerNode.play()
            self.tracks.forEach { t in
                t.playerNode.play()
            }
        }catch{
            print("Error on resuming recording : ",error.localizedDescription)
        }
    }

    @objc func setExternalAudioVolume(volume v:Float) {
        self.playerNode.volume = v
    }
    
    @objc func setMicVolume(volume v:Float) {
        self.engine.inputNode.volume = v
    }
    @objc func setTrackVolume(volume v:Float) {
        self.tracks[selectedTrack].playerNode.volume = v
    }
    
    @objc func toggleReverb() {
        if let node = self.tracks[selectedTrack].effectNode {
            engine.detach(node)
            self.tracks[selectedTrack].effectNode = nil
            if node is AVAudioUnitReverb {
                return
            }
        }
            let rvb = AVAudioUnitReverb()
            do{
            try rvb.loadPreset(at: Bundle.main.url(forResource: "Drums", withExtension: "aupreset")!)
            }catch{
                print("preset erro = ",error.localizedDescription)
            }
            rvb.wetDryMix = 40
            rvb.bypass = false
            rvb.loadFactoryPreset(.largeRoom)
           insertNode(node: rvb)
    }
    
    @objc func setReverbDryWetMix(value:Float){
        (self.tracks[selectedTrack].effectNode as? AVAudioUnitReverb)?.wetDryMix = value
    }
    
    @objc func setReverbPreset(index:Int){
        (self.tracks[selectedTrack].effectNode as? AVAudioUnitReverb)?.loadFactoryPreset(preverbPresets[index])
    }
    
    @objc func addDelay() {
        if let node = self.tracks[selectedTrack].effectNode {
            engine.detach(node)
            self.tracks[selectedTrack].effectNode = nil
            if node is AVAudioUnitDelay {
                return
            }
        }
            let eq = AVAudioUnitDelay()
            eq.feedback = 10
//            eq.lowPassCutoff = 1000
            eq.wetDryMix = 80.0
            eq.delayTime = 0.0
            eq.bypass = false
            self.insertNode(node: eq)
        }
    @objc func setDelayDryWetMix(value:Float){
        (self.tracks[selectedTrack].effectNode as? AVAudioUnitDelay)?.wetDryMix = value
    }
    @objc func setDelayFeedback(value:Int){
        (self.tracks[selectedTrack].effectNode as? AVAudioUnitDelay)?.feedback = Float(value)
    }
    @objc func setDelayTime(sec:Double){
        (self.tracks[selectedTrack].effectNode as? AVAudioUnitDelay)?.delayTime = sec
    }
    
    @objc func addAutoTune() {
        if let node = self.tracks[selectedTrack].effectNode {
            engine.detach(node)
            self.tracks[selectedTrack].effectNode = nil
            if node is AVAudioUnitTimePitch {
                return
            }
        }
            let eq = AVAudioUnitTimePitch()
            eq.pitch = 0
            eq.overlap = 32
            eq.bypass = false
            
            let isPlaying = tracks[selectedTrack].playerNode.isPlaying
            if isPlaying{
                tracks[selectedTrack].playerNode.pause()
            }
            let f = inputMixer.inputFormat(forBus: 0)
            
            self.engine.disconnectNodeOutput(tracks[selectedTrack].playerNode)
            self.engine.disconnectNodeInput(inputMixer)
            self.engine.attach(eq)
        
            self.engine.connect(tracks[selectedTrack].playerNode, to: eq, format: f)
            self.engine.connect(eq, to: inputMixer, format: f)
            
            self.tracks[selectedTrack].effectNode = eq
            if isPlaying{
                tracks[selectedTrack].playerNode.play()
            }
    }
    
    @objc func setAutoTunePitch(value:Float){
        (self.tracks[selectedTrack].effectNode as? AVAudioUnitTimePitch)?.pitch = value
    }
    
    @objc func addEQ() {
        if let node = self.tracks[selectedTrack].effectNode {
            engine.detach(node)
            self.tracks[selectedTrack].effectNode = nil
            if node is AVAudioUnitEQ {
                return
            }
        }
            let eq = AVAudioUnitEQ(numberOfBands: 3)
//            eq.globalGain = 12
//            eq.bands[0].filterType = .highPass
//            eq.bands[1].filterType = .highPass
//            eq.bands[2].filterType = .highPass
            
            eq.bands[0].frequency = 7000
            eq.bands[1].frequency = 1000
            eq.bands[2].frequency = 100
//            eq.bands[0].gain = 3
//            eq.bands[0].filterType = .resonantLowPass
//            eq.bands[0].frequency = 1000
            eq.bypass = false
            
            eq.bands[0].bypass = false
            eq.bands[1].bypass = false
            eq.bands[2].bypass = false
        
            self.insertNode(node: eq)
    }
    @objc func setEQBandgain(bandIndex:Int,decible:Float){
        (self.tracks[selectedTrack].effectNode as? AVAudioUnitEQ)?.bands[bandIndex].gain = decible
    }
    
    private func insertNode(node n:AVAudioUnit){
        let isPlaying = tracks[selectedTrack].playerNode.isPlaying
        if isPlaying{
            tracks[selectedTrack].playerNode.pause()
        }
        let f = inputMixer.inputFormat(forBus: 0)
        if let i = tracks.firstIndex(where: {$0.effectNode != nil}),let node = tracks[i].effectNode{
            engine.detach(node)
            tracks[i].effectNode = nil
        }
        self.engine.disconnectNodeOutput(inputMixer)
        self.engine.disconnectNodeInput(outputMixer)
        self.engine.attach(n)
    
        self.engine.connect(inputMixer, to: n, format: f)
        self.engine.connect(n, to: outputMixer, format: f)
        
        self.tracks[selectedTrack].effectNode = n
        if isPlaying{
            tracks[selectedTrack].playerNode.play()
        }
    }
    
    @objc func uiUpdaterHandler(t time:Timer) {
        currentPosition += 0.1
        if currentPosition < 60{
            self.sendEvent(withName: "TimeChange", body: currentPosition)
        }else{
            stopTimer()
            if self.engine.inputNode.volume == 1{
                self.finishRecording()
            }else{
                self.stopPlaying()
            }
            self.sendEvent(withName: "FinishPlaying", body: nil)
        }
    }
    
    @objc func createWaveformSginalPositive(audioFile:String,_ resolve:RCTPromiseResolveBlock,_ reject:RCTPromiseRejectBlock){
        guard let file = try? AVAudioFile(forReading: .init(fileURLWithPath: audioFile)) else{
            reject("0","File Not Found on path : \(audioFile)",nil)
            return
        }
        let pcmBuffer = AVAudioPCMBuffer(pcmFormat: file.processingFormat, frameCapacity: AVAudioFrameCount(file.length))!
        try! file.read(into: pcmBuffer)

        var signals = [Float]()
        let sampleTime = (pcmBuffer.format.sampleRate/1000)*300
        
        for i in stride(from: 0, to: Int(pcmBuffer.frameLength), by: Int.Stride(sampleTime)) {
            let v = abs(pcmBuffer.floatChannelData![0][i])  //(pcmBuffer.floatChannelData![0][i] + 1)/2
            let dv = powf(10.0, 3)
            let nv = (v * dv).rounded() / dv
            signals.append(nv)
        }
        resolve(signals)
    }
    @objc func createWaveformSginal(audioFile:String, resolve:RCTPromiseResolveBlock, reject:RCTPromiseRejectBlock){
        resolve([23.0])
//        guard let file = try? AVAudioFile(forReading: .init(fileURLWithPath: audioFile)) else{
//            reject("0","File Not Found on path : \(audioFile)",nil)
//            return
//        }
//        let pcmBuffer = AVAudioPCMBuffer(pcmFormat: file.processingFormat, frameCapacity: AVAudioFrameCount(file.length))!
//        try! file.read(into: pcmBuffer)
//
//
////        let sampleTime = (pcmBuffer.format.sampleRate/1000)*300
//
//        let signals:[Float] = Array(UnsafeBufferPointer(start: pcmBuffer.floatChannelData![0], count: Int(pcmBuffer.frameLength)))
//        resolve(signals)
    }
}
