import type { EmitterSubscription } from 'react-native';
import type { Double, Float } from 'react-native/Libraries/Types/CodegenTypes';
/**
 * Create instance of `AudioManager` class to start audio engine services
 * This Audio manager provide below featuers
 * -> Record single track audio
 * -> Import external audio file to audio session for playing while recording
 * -> add multiple track to the session
 * -> apply effect on each audio track indivdually
 */
export declare class AudioManager {
    /**
     * @var audioFrequencyEmitter hold reference of event
     * @var playingFinishEmitter hold reference of event
     * @var timeChangeEmitter hold reference of event
     */
    audioFrequencyEmitter: EmitterSubscription | undefined;
    playingFinishEmitter: EmitterSubscription | undefined;
    timeChangeEmitter: EmitterSubscription | undefined;
    /**
    * use this method to remove events emiter when component getting unmount
    */
    closeEngine(): void;
    /**
     * start audio recording over current selected track if more than one added
     * by default one track alreay added with index 0
     * use track index to modify a track
     * make sure audio engine not recording or playing audio
     */
    startRecording(): void;
    /**
    * stop recording and save for current selected track
    */
    finishRecording(): void;
    /**
     * Play audio session for current postion
     */
    playAudio(): void;
    /**
     * Stop play audio session on current postion
     */
    stopPlaying(): void;
    /**
     * Pause current audio session in playing or stopped state
     * call before seeking audio position
     */
    pause(): void;
    /**
     * Call after when audio session is pause state
     */
    resume(): void;
    /**
     * set external audio path, this music will play during recording and playing the engine
     *
     * @param path {String} - path of audio file
     */
    setExternalAudio(path: String): void;
    /**
    * set external audio volume
    * @param volume {Float} - set external audio volume , renge : 0-1
    */
    setExternalAudioVolume(volume: Float): void;
    /**
     * set input mic recording volume
     * @param volume {Float} - set volume range 0-1
     */
    setMicVolume(volume: Float): void;
    /**
     * Add an audio track to current audio session for recording
     * Make sure audio engine is not running
     */
    addTrack(): void;
    /**
     * Set current selected track volume
     * @param volume {Float} - set volume range 0-1
     */
    setTrackVolume(volume: Float): void;
    /**
     * remove an audio track to current audio session
     * @param trackIndex {Int} pointing a track in audio session
     */
    removeTrack(trackIndex: number): void;
    /**
     * Select a audio track in current audio session
     * To record over or applying effect on it
     * @param trackIndex {Int} pointing a track in audio session
     */
    selectTrack(trackIndex: number): void;
    /**
     * change current audio engine position for recording or playing audio
     * @param seconds {Double} set position must be less than maximum position
     */
    seekPlayer(seconds: Double): void;
    /**
     * Generate waveform signal for an audio file
     * @param path {String} audio file path
     * @returns float array
     */
    createWaveformSginal(filePath: String): Promise<[Float]>;
    /**
     * Generate waveform positive signal for an audio file
     * @param path {String} audio file path
     * @returns float array
     */
    createWaveformSginalPositive(filePath: String): Promise<[Float]>;
    /**
     * add event to listen updates of audio engine current position
     * @param onTimeChange {(p:Double)=>void} called when enging is running for recording for playing audio
     */
    setOnTimeChange(onTimeChange: (p: Double) => void): void;
    /**
      * add event listener to detect audio playing finishing
      * @param onFinishPlaying {()=>void} called when audio playing finished
      */
    setOnFinishPlaying(onFinishPlaying: () => void): void;
    /**
     * add event to get audio recording mic frequncy after 20 milliseconds
     * @param audioFrequencyListener {Float)=>void} method fired when signal received from mic
     * this represent the frequency of current buffer received from mic
     */
    setAudioFrequencyListener(audioFrequencyListener: (p: Float) => void): void;
    /**
     * Use this method to apply or remove effect
     * if effect already addded then it will be removed from selected track
     */
    addDelay(): void;
    /**
     *   Blend of the wet and dry signals
     *   @param value {Float}
     *   Range:      0 (all dry) -> 100 (all wet)
     *   Default:    100
     *   Unit:       Percent
     */
    setDelayDryWetMix(value: Float): void;
    /**
     *   Amount of the output signal fed back into the delay line
     *   @param value {Int} set value
     *    Range:      -100 -> 100
     *    Default:    5
     *    Unit:       Percent
     */
    setFeedback(value: number): void;
    /**
    * Time taken by the delayed input signal to reach the output
    * @param seconds {Double}
    *    Range:      0 -> 2
    *    Default:    1
    *    Unit:       Seconds
   */
    setDelayTime(seconds: Double): void;
    /**
      * Use this method to apply or remove effect
      * if effect already addded then it will be removed from selected track
      */
    addAutoTune(): void;
    /**
     * amount by which the input signal is pitch shifted
     * 1 octave  = 1200 cents
     * 1 musical semitone  = 100 cents
     * @param value {Float}
     *    Range:      -2400 -> 2400
     *    Default:    0.0
     *    Unit:       Cents
    */
    setAutoTunePitch(value: Float): void;
    /**
    * Use this method to apply or remove effect
    * if effect already addded then it will be removed from selected track
    */
    addEQ(): void;
    /**
     *  gain in Hertz.
     * @param bandIndex {Int} set band index 0, 1, 2
     * @param frequency {Float}
     *Range:      -96 -> 24
     *Default:    0
     *Unit:       dB
     */
    setEQBandgain(bandIndex: number, db: Float): void;
    /**
    * Use this method to apply or remove effect
    * if effect already addded then it will be removed from selected track
    */
    toggleReverb(): void;
    /**
      *  Frequency in Hertz.
      * @param bandIndex {Int} set band index 0, 1, 2
      * @param frequency {Float}
      * Range:      20 -> (SampleRate/2)
      * SampleRate commonly used is 48000 Hz
      * Unit:       Hertz
      */
    setReverbDryWetMix(value: Float): void;
    /**
     * load a reverb preset
     * @param presetIndex {Int} preset index:
     *  Preset index description
     * smallRoom = 0
     * mediumRoom = 1
     * largeRoom = 2
     * mediumHall = 3
     * largeHall = 4
     * plate = 5
     * mediumChamber = 6
     * largeChamber = 7
     * cathedral = 8
     * largeRoom2 = 9
     * mediumHall2 = 10
     * mediumHall3 = 11
     * largeHall2 = 12
     *   Default:    mediumRoom : 1
  */
    setReverbPreset(presetIndex: number): void;
    /**
     * use to export output file with .aac file format
     * @returns return Result with exported audio file path if successfully rendered else return nil
     */
    exportFile(): Promise<String>;
}
//# sourceMappingURL=index.d.ts.map