
import type { EmitterSubscription } from 'react-native';
import { NativeEventEmitter } from 'react-native';
import { NativeModules, Platform } from 'react-native';
import type { Double, Float } from 'react-native/Libraries/Types/CodegenTypes';

const LINKING_ERROR =
  `The package 'react-native-native-audio-engine' doesn't seem to be linked. Make sure: \n\n` +
  Platform.select({ ios: "- You have run 'pod install'\n", default: '' }) +
  '- You rebuilt the app after installing the package\n' +
  '- You are not using Expo Go\n';

const NativeAudioEngine = NativeModules.NativeAudioEngine
  ? NativeModules.NativeAudioEngine
  : new Proxy(
      {},
      {
        get() {
          throw new Error(LINKING_ERROR);
        },
      }
    );

// export function multiply(a: number, b: number): Promise<number> {
//   return NativeAudioEngine.multiply(a, b);
// }

  /**
   * Create instance of `AudioManager` class to start audio engine services
   * This Audio manager provide below featuers
   * -> Record single track audio
   * -> Import external audio file to audio session for playing while recording 
   * -> add multiple track to the session
   * -> apply effect on each audio track indivdually
   */
export class AudioManager{
  /**
   * @var audioFrequencyEmitter hold reference of event 
   * @var playingFinishEmitter hold reference of event 
   * @var timeChangeEmitter hold reference of event 
   */
  audioFrequencyEmitter: EmitterSubscription | undefined= undefined;  
  playingFinishEmitter: EmitterSubscription | undefined= undefined;
  timeChangeEmitter: EmitterSubscription | undefined= undefined;
  
   /**
   * use this method to remove events emiter when component getting unmount 
   */
  closeEngine(){
    this.audioFrequencyEmitter?.remove();
    this.playingFinishEmitter?.remove();
    this.timeChangeEmitter?.remove();
  }

 /**
  * start audio recording over current selected track if more than one added
  * by default one track alreay added with index 0
  * use track index to modify a track
  * make sure audio engine not recording or playing audio
  */
  startRecording(){
    NativeAudioEngine.startRecording()
  }

  /**
  * stop recording and save for current selected track
  */
  finishRecording(){
    NativeAudioEngine.finishRecording()
  }

  /**
   * Play audio session for current postion 
   */
  playAudio(){
    NativeAudioEngine.playAudio()
  }
  
  /**
   * Stop play audio session on current postion 
   */
  stopPlaying(){
    NativeAudioEngine.stopPlaying()
  }

  /**
   * Pause current audio session in playing or stopped state
   * call before seeking audio position
   */
  pause(){
    NativeAudioEngine.pause()
  }

  /**
   * Call after when audio session is pause state
   */
  resume(){
    NativeAudioEngine.resume()
  }

  /**
   * set external audio path, this music will play during recording and playing the engine
   *
   * @param path {String} - path of audio file
   */
  setExternalAudio(path:String){
    NativeAudioEngine.setExternalAudio(path)
  }

   /**
   * set external audio volume
   * @param volume {Float} - set external audio volume , renge : 0-1
   */
  setExternalAudioVolume(volume:Float){
    NativeAudioEngine.setExternalAudioVolume(volume);
  }

  /**
   * set input mic recording volume 
   * @param volume {Float} - set volume range 0-1
   */
  setMicVolume(volume:Float){
    NativeAudioEngine.setMicVolume(volume);
  }

  /**
   * Add an audio track to current audio session for recording
   * Make sure audio engine is not running
   */
  addTrack(){
    NativeAudioEngine.addTrack()
  }

  /**
   * Set current selected track volume
   * @param volume {Float} - set volume range 0-1
   */
  setTrackVolume(volume:Float){
    NativeAudioEngine.setTrackVolume(volume)
  }

  /**
   * remove an audio track to current audio session
   * @param trackIndex {Int} pointing a track in audio session
   */
  removeTrack(trackIndex:number){
    NativeAudioEngine.removeTrack(trackIndex)
  }

  /**
   * Select a audio track in current audio session
   * To record over or applying effect on it
   * @param trackIndex {Int} pointing a track in audio session
   */
  selectTrack(trackIndex:number){
    NativeAudioEngine.selectTrack(trackIndex)
  }

  /**
   * change current audio engine position for recording or playing audio
   * @param seconds {Double} set position must be less than maximum position
   */
  seekPlayer(seconds:Double){
    NativeAudioEngine.seekPlayer(seconds)
  }

  /**
   * Generate waveform signal for an audio file
   * @param path {String} audio file path 
   * @returns float array 
   */
  createWaveformSginal(filePath:String):Promise<[Float]>{
    return NativeAudioEngine.createWaveformSginal(filePath)
  }


  /**
   * Generate waveform positive signal for an audio file
   * @param path {String} audio file path 
   * @returns float array 
   */
  createWaveformSginalPositive(filePath:String):Promise<[Float]>{
    return NativeAudioEngine.createWaveformSginalPositive(filePath)
  }

  /**
   * add event to listen updates of audio engine current position 
   * @param onTimeChange {(p:Double)=>void} called when enging is running for recording for playing audio
   */
  setOnTimeChange(onTimeChange:(p:Double)=>void){
    this.playingFinishEmitter = new NativeEventEmitter(NativeAudioEngine).addListener(
      'TimeChange',
      (param) => onTimeChange(param)
    );
  }
 /**
   * add event listener to detect audio playing finishing
   * @param onFinishPlaying {()=>void} called when audio playing finished
   */
  setOnFinishPlaying(onFinishPlaying:()=>void){
    this.timeChangeEmitter = new NativeEventEmitter(NativeAudioEngine).addListener(
      'FinishPlaying',
      (_param) => onFinishPlaying()
    );
  }
  /**
   * add event to get audio recording mic frequncy after 20 milliseconds
   * @param audioFrequencyListener {Float)=>void} method fired when signal received from mic
   * this represent the frequency of current buffer received from mic
   */
  setAudioFrequencyListener(audioFrequencyListener:(p:Float)=>void){
    this.audioFrequencyEmitter = new NativeEventEmitter(NativeAudioEngine).addListener(
      'FrequencyEmitter',
      (reminder) => audioFrequencyListener(reminder)
    );
  }


  //Effects Section
  /**
   * Use this method to apply or remove effect
   * if effect already addded then it will be removed from selected track
   */
  addDelay(){
    NativeAudioEngine.addDelay()
  }

   /** 
    *   Blend of the wet and dry signals
    *   @param value {Float}
    *   Range:      0 (all dry) -> 100 (all wet)
    *   Default:    100
    *   Unit:       Percent
    */
  setDelayDryWetMix(value:Float){
    NativeAudioEngine.setDelayDryWetMix(value)
  }
  /**
   *   Amount of the output signal fed back into the delay line
   *   @param value {Int} set value
   *    Range:      -100 -> 100
   *    Default:    5
   *    Unit:       Percent
   */
  setFeedback(value:number){
    NativeAudioEngine.setFeedback(value)
  }

   /** 
   * Time taken by the delayed input signal to reach the output
   * @param seconds {Double}
   *    Range:      0 -> 2
   *    Default:    1
   *    Unit:       Seconds
  */
  setDelayTime(seconds:Double){
    NativeAudioEngine.setDelayTime(seconds)
  }

 /**
   * Use this method to apply or remove effect
   * if effect already addded then it will be removed from selected track
   */
  addAutoTune(){
    NativeAudioEngine.addAutoTune()
  }

  /** 
   * amount by which the input signal is pitch shifted
   * 1 octave  = 1200 cents
   * 1 musical semitone  = 100 cents
   * @param value {Float}
   *    Range:      -2400 -> 2400
   *    Default:    0.0
   *    Unit:       Cents
  */
  setAutoTunePitch(value:Float){
    NativeAudioEngine.setAutoTunePitch(value)
  }

   /**
   * Use this method to apply or remove effect
   * if effect already addded then it will be removed from selected track
   */
  addEQ(){
    NativeAudioEngine.addEQ()
  }

  /** 
   *  gain in Hertz.
   * @param bandIndex {Int} set band index 0, 1, 2
   * @param frequency {Float}
   *Range:      -96 -> 24
   *Default:    0
   *Unit:       dB
   */
  setEQBandgain(bandIndex:number,db:Float){
    NativeAudioEngine.setEQBandgain(bandIndex,db)
  }

   /**
   * Use this method to apply or remove effect
   * if effect already addded then it will be removed from selected track
   */
  toggleReverb(){
    NativeAudioEngine.toggleReverb()
  }
 /** 
   *  Frequency in Hertz.
   * @param bandIndex {Int} set band index 0, 1, 2
   * @param frequency {Float}
   * Range:      20 -> (SampleRate/2)
   * SampleRate commonly used is 48000 Hz
   * Unit:       Hertz
   */
  setReverbDryWetMix(value:Float){
    NativeAudioEngine.setReverbDryWetMix(value)
  }
  /** 
   * load a reverb preset
   * @param presetIndex {Int} preset index:
   *  Preset index description
   * smallRoom = 0
   * mediumRoom = 1
   * largeRoom = 2
   * mediumHall = 3
   * largeHall = 4
   * plate = 5
   * mediumChamber = 6
   * largeChamber = 7
   * cathedral = 8
   * largeRoom2 = 9
   * mediumHall2 = 10
   * mediumHall3 = 11
   * largeHall2 = 12
   *   Default:    mediumRoom : 1
*/
  setReverbPreset(presetIndex:number){
    NativeAudioEngine.setReverbPreset(presetIndex)
  }

  /**
   * use to export output file with .aac file format 
   * @returns return Result with exported audio file path if successfully rendered else return nil
   */
  exportFile():Promise<String>{
    return NativeAudioEngine.exportFile()
  }
}